function [ux, uy, energies] = update_flow(i0, i1w, ux, uy, lamb, nbiter)
    %update_flow: main function that updates the optical flow for each step
    h = size(i0, 1); 
    w = size(i0, 2);
    nab = nabla(h, w);
    s = 3; t = 3;  % s * t should be greater than upperbound (8 for nabla)
    % build up vectorized quantities
    I0 = i0(:); 
    I1w = i1w(:);
    DI1w = D(I1w, h, w);
    diffI = I1w - I0;
    U = cat(2, ux(:), uy(:));
    U_bar = U;
    P = zeros(2 * h * w, 2);
    energies = zeros(nbiter, 1);
    % perform PDHG updates
    for i = 1:nbiter
        U_old = U;
        U = update_U(U, U_bar, P, nab, diffI, DI1w, s, lamb);
        P = update_P(2*U-U_old, P, nab, t);
        energies(i) = compute_primal_energy(...
            U, U_bar, nab, diffI, DI1w, lamb);
    end
    ux = reshape(U(:, 1), [h, w]); 
    uy = reshape(U(:, 2), [h, w]);
end


function [dx] = Dx(h, w)
    %Dx: 
    e = ones(w, 1);
    Dx_tilde = spdiags([e -e], -1:0, w, w);
    Dx_tilde(end, end) = 0;
    dx = kron(Dx_tilde', speye(h));
end


function [dy] = Dy(h, w)
    e = ones(h, 1);
    Dy_tilde = spdiags([-e e], 0:1, h, h);
    Dy_tilde(end, end) = 0;
    dy = kron(speye(w), Dy_tilde);
end

   
function [nab] = nabla(h, w)
    %nabla: build (sparse) nabla operator
    nab = cat(1, Dx(h, w),  Dy(h, w));
end


function [dI] = D(I, h, w)
%D: compute image gradient along x,y directions. return shape (h*w, 2)
    dI = cat(2, Dx(h, w) * I,  Dy(h, w) * I);
end


function [r] = rho(U, U_bar, nabP_s, diffI, DI1w, lamb_s)
    numerator = sum(DI1w .* (U - U_bar - nabP_s), 2) + diffI;
    denominator = lamb_s * (vecnorm(DI1w, 2, 2).^2) + 1e-20;
    r = numerator ./ denominator;
end


function [U_new] = update_U(U, U_bar, P, nab, diffI, DI1w, s, lamb)
    nabP_s = nab' * P / s;
    lamb_s = lamb / s;
    rho_U = rho(U, U_bar, nabP_s, diffI, DI1w, lamb_s);
    U_new = U - nabP_s - lamb_s * DI1w .* min(max(rho_U, -1), 1);
end


function [pm] = proj_C(m)
    %proj_C: row-wise projection to unit l2-ball
    m_norm = vecnorm(m, 2, 2);  % row-wise l2 norm
    m_norm(m_norm <= 1, :) = 1.0;
    pm = m ./ m_norm;
end


function [P_new] = update_P(U_extrapolate, P, nab, t)
    P_new = proj_C(P + nab * U_extrapolate / t);
end


function [eng] = compute_primal_energy(U, U_bar, nab, diffI, DI1w, lamb)
    eng = lamb * sum(abs(sum(DI1w .* (U - U_bar), 2) + diffI)) + ...
        sum(vecnorm(nab * U, 2, 2));
end
